#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "pktlib.h"
#include "lib.h"

static struct {
  pktif_base_t list;
#ifndef USE_MINGW
  int fdnum;
  fd_set readfds; /* ץ󤷤󥿡եΰ */
  fd_set fds; /* select()ˤä֤줿read()ǽ󥿡ե */
#endif
} pktif_list; /* ץ󤷤󥿡եΥꥹ */

char *pktif_get_name(pktif_t pktif)
{
  return ((pktif_base_t)pktif)->name;
}

int pktif_get_fd(pktif_t pktif)
{
  return ((pktif_base_t)pktif)->fd;
}

unsigned long pktif_get_flags(pktif_t pktif)
{
  return ((pktif_base_t)pktif)->flags;
}

int pktif_get_linktype(pktif_t pktif)
{
  return ((pktif_base_t)pktif)->linktype;
}

int pktif_get_buffer_size(pktif_t pktif)
{
  return ((pktif_base_t)pktif)->buffer_size;
}

int pktif_get_send_count(pktif_t pktif)
{
  return ((pktif_base_t)pktif)->send_count;
}

int pktif_get_recv_count(pktif_t pktif)
{
  return ((pktif_base_t)pktif)->recv_count;
}

void *pktif_get_option(pktif_t pktif)
{
  return ((pktif_base_t)pktif)->option;
}

pktif_t pktif_get_next(pktif_t pktif)
{
  return (pktif_t)((pktif_base_t)pktif)->next;
}

int pktlib_linktype_offset(int linktype)
{
  switch (linktype) {
#ifdef __linux__
  case DLT_NULL: /* fall through */
#endif
  case DLT_EN10MB: return 2;
  default:
    break;
  }
  return 0;
}

int pktif_get_default_offset(pktif_t pktif)
{
  return pktlib_linktype_offset(pktif_get_linktype(pktif));
}

static void init()
{
  static int initialized = 0;
  if (!initialized) {
    pktif_list.list = NULL;
#ifndef USE_MINGW
    pktif_list.fdnum = 0;
    FD_ZERO(&pktif_list.readfds);
    FD_ZERO(&pktif_list.fds);
#endif
    initialized = 1;
  }
}

pktif_t pktlib_iflist_get_list()
{
  init();
  return (pktif_t)pktif_list.list;
}

#ifndef USE_MINGW
int pktlib_iflist_fd_set(int fd)
{
  init();
  if (pktif_list.fdnum < fd + 1)
    pktif_list.fdnum = fd + 1;
  FD_SET(fd, &pktif_list.readfds);
  return 0;
}

int pktlib_iflist_fd_clr(int fd)
{
  init();
  FD_CLR(fd, &pktif_list.fds);
  return 0;
}

int pktlib_iflist_fd_isset(int fd)
{
  init();
  return FD_ISSET(fd, &pktif_list.fds);
}
#endif

int pktlib_iflist_set(pktif_t pktif)
{
  pktif_base_t base = (pktif_base_t)pktif;

  init();

  if (base->flags & PKTIF_OPEN_FLAG_SELECT_DISABLE)
    return -1;

  base->next = pktif_list.list; /* 󥯥ꥹȤ³ */
  pktif_list.list = base;

  if (base->flags & PKTIF_OPEN_FLAG_SELECT_NOSELECT)
    return 0;

#ifndef USE_MINGW
  pktlib_iflist_fd_set(base->fd); /* 󥿡եɲ */
#endif
  return 1;
}

#ifndef USE_MINGW
pktif_t pktlib_iflist_select(int usec)
{
  pktif_t pktif;
  pktif_base_t base;
  struct timeval t, *tp = NULL;

  init();

  /* ĤäƤѥåȤ̵Ĵ٤ */
  for (base = pktif_list.list; base; base = base->next) {
    if (base->flags & PKTIF_OPEN_FLAG_SELECT_NOSELECT)
      continue;
    pktif = (pktif_t)base;
    if (!pktif_is_empty(pktif)) /* Хåե˻ĤäƤ */
      return pktif;
    if (pktlib_iflist_fd_isset(base->fd)) { /* μԤ */
      pktlib_iflist_fd_clr(base->fd);
      return pktif;
    }
  }

  if (usec >= 0) { /* ॢȻ֤(ͤʤ饿ॢȤʤ) */
    t.tv_sec  = usec / 1000000;
    t.tv_usec = usec % 1000000;
    tp = &t;
  }
  pktif_list.fds = pktif_list.readfds;
  pktif = NULL;
  if (select(pktif_list.fdnum, &pktif_list.fds, NULL, NULL, tp) > 0) {
    for (base = pktif_list.list; base; base = base->next) {
      if (pktlib_iflist_fd_isset(base->fd)) { /* Ԥ */
	pktif = (pktif_t)base;
	pktlib_iflist_fd_clr(base->fd);
	break;
      }
    }
  }
  return pktif;
}
#endif

void pktlib_error_exit(char *message)
{
  fprintf(stderr, "%s", message);
  exit(1);
}
